@@ -44,6 +44,11 @@ struct _GHexWindow
     BonoboUIComponent *uic;
     gboolean changed, undo_sens, redo_sens;
 
+    HexDialog *dialog;
+    GtkWidget *dialog_widget;
+
+    AdvancedFindDialog *advanced_find_dialog;
+
     gchar **uris_to_open;
 };
 
